@@ -1,7 +1,7 @@
 # UpdatesAvailable.py
 # -*- Mode: Python; indent-tabs-mode: nil; tab-width: 4; coding: utf-8 -*-
 #
-#  Copyright (c) 2004-2012 Canonical
+#  Copyright (c) 2004-2013 Canonical
 #                2004 Michiel Sikkes
 #                2005 Martin Willemoes Hansen
 #                2010 Mohamed Amine IL Idrissi
@@ -12,6 +12,7 @@
 #          Mohamed Amine IL Idrissi <ilidrissiamine@gmail.com>
 #          Alex Launi <alex.launi@canonical.com>
 #          Michael Terry <michael.terry@canonical.com>
+#          Dylan McCall <dylanmccall@ubuntu.com>
 #
 #  This program is free software; you can redistribute it and/or
 #  modify it under the terms of the GNU General Public License as
@@ -30,6 +31,7 @@
 
 from __future__ import absolute_import, print_function
 
+from gi.repository import GLib
 from gi.repository import Gtk
 from gi.repository import Gdk
 from gi.repository import GObject
@@ -46,23 +48,22 @@ import apt_pkg
 import os
 import re
 import logging
-import operator
 import subprocess
 import time
 import threading
-import xml.sax.saxutils
 
 from gettext import gettext as _
 from gettext import ngettext
 
 
-from .Core.utils import humanize_size
+from .Core.utils import (get_package_label, humanize_size)
 from .Core.AlertWatcher import AlertWatcher
+from .Core.UpdateList import UpdateSystemGroup
 
 from DistUpgrade.DistUpgradeCache import NotEnoughFreeSpaceError
 
 from .ChangelogViewer import ChangelogViewer
-from .SimpleGtk3builderApp import SimpleGtkbuilderApp
+from .SimpleGtkbuilderApp import SimpleGtkbuilderApp
 from .UnitySupport import UnitySupport
 
 
@@ -70,15 +71,138 @@ from .UnitySupport import UnitySupport
 
 # FIXME:
 # - kill "all_changes" and move the changes into the "Update" class
+# - screen reader does not read update toggle state
+# - screen reader does not say "Downloaded" for downloaded updates
 
 # list constants
-(LIST_CONTENTS, LIST_NAME, LIST_PKG,
- LIST_ORIGIN, LIST_TOGGLE_CHECKED) = range(5)
+(LIST_NAME, LIST_UPDATE_DATA, LIST_SIZE, LIST_TOGGLE_ACTIVE) = range(4)
 
 # NetworkManager enums
 from .Core.roam import NetworkManagerHelper
 
 
+class UpdateData():
+    def __init__(self, groups, group, item):
+        self.groups = groups if groups else []
+        self.group = group
+        self.item = item
+
+
+class CellAreaPackage(Gtk.CellAreaBox):
+    """This CellArea lays our package cells side by side, without allocating
+       width for a cell if it isn't present (like icons for header labels).
+       We assume that the last cell should be expanded to fill remaining space,
+       and other cells have a fixed width.
+    """
+
+    def __init__(self, indent_toplevel=False):
+        Gtk.CellAreaBox.__init__(self)
+        self.indent_toplevel = indent_toplevel
+        self.column = None
+        self.cached_cell_size = {}
+
+    def do_foreach_alloc(self, context, widget, cell_area_in, bg_area_in,
+                         callback):
+        cells = []
+        def gather(cell, data):
+            cells.append(cell)
+        self.foreach(gather, None)
+
+        cell_is_hidden = {}
+
+        # Record the space required by each cell
+        for cell_number, cell in enumerate(cells):
+            # Detect if this cell should be allocated space
+            if isinstance(cell, Gtk.CellRendererPixbuf):
+                gicon = cell.get_property("gicon")
+                hide_cell = gicon is None
+            else:
+                hide_cell = False
+            cell_is_hidden[cell_number] = hide_cell
+
+            if not hide_cell and cell_number not in self.cached_cell_size:
+                min_size, natural_size = cell.get_preferred_width(widget)
+                self.cached_cell_size[cell_number] = natural_size
+
+        cell_area = cell_area_in.copy()
+        bg_area = bg_area_in.copy()
+        spacing = self.get_property("spacing")
+        cell_start = self.get_cell_start(widget)
+        orig_end = cell_area.width + cell_area.x
+        cur_path = self.get_current_path_string()
+        depth = Gtk.TreePath.new_from_string(cur_path).get_depth()
+
+        # And finally, start handling each cell
+        extra_cell_width = 0
+        cell_area.x = cell_start
+        cell_area.width = 0
+
+        last_cell_number = len(cells) - 1
+        for cell_number, cell in enumerate(cells):
+            is_last_cell = cell_number == last_cell_number
+            cell_size = self.cached_cell_size.get(cell_number, 0)
+
+            if cell_area.width > 0 and extra_cell_width == 0:
+                cell_area.x += cell_area.width + spacing
+
+            if cell_number == 0:
+                # The first cell is affected by its depth in the tree
+                if not cell_is_hidden[1] and self.indent_toplevel:
+                    # if not a header, align with header rows
+                    depth += 1
+                if depth > 1:
+                    indent = max(0, depth - 1)
+                    indent_size = cell_size * indent
+                    if depth == 2:
+                        indent_extra = spacing
+                    elif depth == 3:
+                        indent_extra = spacing + 1
+                    else:
+                        indent_extra = spacing * indent
+                    cell_area.x += indent_size + indent_extra
+
+            if is_last_cell:
+                cell_size = max(cell_size, orig_end - cell_area.x)
+            if not cell_is_hidden[cell_number]:
+                cell_area.width = cell_size + extra_cell_width
+                extra_cell_width = 0
+            else:
+                cell_area.width = 0
+                extra_cell_width = cell_size + spacing
+
+            if callback(cell, cell_area.copy(), bg_area.copy()):
+                return
+
+    def do_event(self, context, widget, event, cell_area, flags):
+        # This override is just to trick our parent implementation into
+        # allowing clicks on toggle cells when they are where the expanders
+        # usually are.  It doesn't expect that, so we expand the cell_area
+        # here to be equivalent to bg_area.
+        cell_start = self.get_cell_start(widget)
+        cell_area.width = cell_area.width + cell_area.x - cell_start
+        cell_area.x = cell_start
+        return Gtk.CellAreaBox.do_event(self, context, widget, event,
+                                        cell_area, flags)
+
+    def get_cell_start(self, widget):
+        if not self.column:
+            return 0
+        else:
+            val = GObject.Value()
+            val.init(int)
+            widget.style_get_property("horizontal-separator", val)
+            h_sep = val.get_int()
+            widget.style_get_property("grid-line-width", val)
+            line_width = val.get_int()
+            cell_start = self.column.get_x_offset() - h_sep - line_width
+            if not self.indent_toplevel:  # i.e. if no headers
+                widget.style_get_property("expander-size", val)
+                spacing = self.get_property("spacing")
+                # Hardcode 4 because GTK+ hardcodes 4 internally
+                cell_start = cell_start + val.get_int() + 4 + spacing
+            return cell_start
+
+
 class UpdatesAvailable(SimpleGtkbuilderApp):
 
     def __init__(self, app, header=None, desc=None):
@@ -134,38 +258,71 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
         self.update_close_button()
 
         # the treeview (move into it's own code!)
-        self.store = Gtk.ListStore(str, str, GObject.TYPE_PYOBJECT,
-                                   GObject.TYPE_PYOBJECT, bool)
+        self.store = Gtk.TreeStore(str, GObject.TYPE_PYOBJECT, str, bool)
         self.treeview_update.set_model(self.store)
-        self.treeview_update.set_headers_clickable(True)
+
+        restart_icon_renderer = Gtk.CellRendererPixbuf()
+        restart_icon_renderer.set_property("xpad", 4)
+        restart_icon_renderer.set_property("ypad", 2)
+        restart_icon_renderer.set_property("stock-size", Gtk.IconSize.MENU)
+        restart_icon_renderer.set_property("follow-state", True)
+        restart_column = Gtk.TreeViewColumn(None, restart_icon_renderer)
+        restart_column.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
+        restart_column.set_fixed_width(20)
+        self.treeview_update.append_column(restart_column)
+        restart_column.set_cell_data_func(restart_icon_renderer,
+                                          self.restart_icon_renderer_data_func)
+
+        pkg_area = CellAreaPackage(bool(self.list.security_groups))
+        pkg_column = Gtk.TreeViewColumn.new_with_area(pkg_area)
+        pkg_area.column = pkg_column
+        pkg_column.set_title(_("Install"))
+        pkg_column.set_property("spacing", 4)
+        pkg_column.set_expand(True)
+        self.treeview_update.append_column(pkg_column)
+
+        pkg_toggle_renderer = Gtk.CellRendererToggle()
+        pkg_toggle_renderer.set_property("ypad", 2)
+        pkg_toggle_renderer.connect("toggled", self.on_update_toggled)
+        pkg_column.pack_start(pkg_toggle_renderer, False)
+        pkg_column.add_attribute(pkg_toggle_renderer,
+                                 'active', LIST_TOGGLE_ACTIVE)
+        pkg_column.set_cell_data_func(pkg_toggle_renderer,
+                                      self.pkg_toggle_renderer_data_func)
+
+        pkg_icon_renderer = Gtk.CellRendererPixbuf()
+        pkg_icon_renderer.set_property("ypad", 2)
+        pkg_icon_renderer.set_property("stock-size", Gtk.IconSize.MENU)
+        pkg_column.pack_start(pkg_icon_renderer, False)
+        pkg_column.set_cell_data_func(pkg_icon_renderer,
+                                      self.pkg_icon_renderer_data_func)
+
+        pkg_label_renderer = Gtk.CellRendererText()
+        pkg_label_renderer.set_property("ypad", 2)
+        pkg_label_renderer.set_property("ellipsize", Pango.EllipsizeMode.END)
+        pkg_column.pack_start(pkg_label_renderer, True)
+        pkg_column.set_cell_data_func(pkg_label_renderer,
+                                      self.pkg_label_renderer_data_func)
+
+        size_renderer = Gtk.CellRendererText()
+        size_renderer.set_property("xpad", 6)
+        size_renderer.set_property("ypad", 0)
+        size_renderer.set_property("xalign", 1)
+        # 1.0/1.2 == PANGO.Scale.SMALL. Constant is not (yet) introspected.
+        size_renderer.set_property("scale", 1.0 / 1.2)
+        size_column = Gtk.TreeViewColumn(_("Download"), size_renderer,
+                                         text=LIST_SIZE)
+        size_column.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
+        self.treeview_update.append_column(size_column)
+
+        self.treeview_update.set_headers_visible(True)
+        self.treeview_update.set_headers_clickable(False)
         self.treeview_update.set_direction(Gtk.TextDirection.LTR)
-
-        tr = Gtk.CellRendererText()
-        tr.set_property("xpad", 6)
-        tr.set_property("ypad", 6)
-        cr = Gtk.CellRendererToggle()
-        cr.set_property("activatable", True)
-        cr.set_property("xpad", 6)
-        cr.connect("toggled", self.toggled)
-
-        column_install = Gtk.TreeViewColumn(_("Install"), cr,
-                                            active=LIST_TOGGLE_CHECKED)
-        column_install.set_cell_data_func(cr, self.install_column_view_func)
-        column = Gtk.TreeViewColumn(_("Name"), tr, markup=LIST_CONTENTS)
-        column.set_resizable(True)
-
-        column_install.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
-        column_install.set_fixed_width(30)
-        column.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
-        column.set_fixed_width(100)
         self.treeview_update.set_fixed_height_mode(False)
-
-        self.treeview_update.append_column(column_install)
-        column_install.set_visible(True)
-        self.treeview_update.append_column(column)
+        self.treeview_update.set_expander_column(pkg_column)
         self.treeview_update.set_search_column(LIST_NAME)
         self.treeview_update.connect("button-press-event",
-                                     self.show_context_menu)
+                                     self.on_treeview_button_press)
 
         # setup the help viewer and disable the help button if there
         # is no viewer available
@@ -210,26 +367,97 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
             self.button_close.set_use_underline(False)
 
     def install_all_updates(self, menu, menuitem, data):
-        self.select_all_updgrades(None)
+        self.select_all_upgrades(None)
         self.on_button_install_clicked(None)
 
-    def install_column_view_func(self, cell_layout, renderer, model, iter,
-                                 data):
-        pkg = model.get_value(iter, LIST_PKG)
-        if pkg is None:
-            renderer.set_property("activatable", True)
-            return
-        current_state = renderer.get_property("active")
-        to_install = pkg.marked_install or pkg.marked_upgrade
-        renderer.set_property("active", to_install)
-        # we need to update the store as well to ensure orca knowns
-        # about state changes (it will not read view_func changes)
-        if to_install != current_state:
-            self.store[iter][LIST_TOGGLE_CHECKED] = to_install
-        if pkg.name in self.list.held_back:
-            renderer.set_property("activatable", False)
+    def restart_icon_renderer_data_func(self, cell_layout, renderer, model,
+                                        iter, data):
+        def pkg_requires_restart(pkg):
+            restart_condition = pkg.candidate.record.get('XB-Restart-Required')
+            return restart_condition == 'system'
+
+        data = model.get_value(iter, LIST_UPDATE_DATA)
+        path = model.get_path(iter)
+
+        requires_restart = False
+        if data.item:
+            requires_restart = pkg_requires_restart(data.item.pkg)
+        elif data.group:
+            if not self.treeview_update.row_expanded(path):
+                # A package in the group requires restart
+                for group_item in data.group.items:
+                    if pkg_requires_restart(group_item.pkg):
+                        requires_restart = True
+                        break
+
+        # FIXME: Non-standard, incorrect icon name (from app category).
+        # Theme support for what we want seems to be lacking.
+        if requires_restart:
+            restart_icon_names = ['view-refresh-symbolic',
+                                  'system-restart',
+                                  'system-reboot']
+            gicon = Gio.ThemedIcon.new_from_names(restart_icon_names)
         else:
-            renderer.set_property("activatable", True)
+            gicon = None
+        renderer.set_property("gicon", gicon)
+
+    def pkg_toggle_renderer_data_func(self, cell_layout, renderer, model, iter,
+                                      data):
+        data = model.get_value(iter, LIST_UPDATE_DATA)
+
+        activatable = False
+        inconsistent = False
+        if data.item:
+            activatable = data.item.pkg.name not in self.list.held_back
+            inconsistent = False
+        elif data.group:
+            activatable = True
+            inconsistent = data.group.selection_is_inconsistent()
+        elif data.groups:
+            activatable = True
+            inconsistent = False
+            saw_install = None
+            for group in data.groups:
+                for item in group.items:
+                    pkg = item.pkg
+                    this_install = pkg.marked_install or pkg.marked_upgrade
+                    if saw_install is not None and saw_install != this_install:
+                        inconsistent = True
+                        break
+                    saw_install = this_install
+                if inconsistent:
+                    break
+
+        # The "active" attribute is already set via LIST_TOGGLE_ACTIVE in the
+        # tree model, so we don't set it here.
+        renderer.set_property("activatable", activatable)
+        renderer.set_property("inconsistent", inconsistent)
+
+    def pkg_icon_renderer_data_func(self, cell_layout, renderer, model, iter,
+                                    data):
+        data = model.get_value(iter, LIST_UPDATE_DATA)
+
+        gicon = None
+        if data.group:
+            gicon = data.group.icon
+        elif data.item:
+            gicon = data.item.icon
+
+        renderer.set_property("gicon", gicon)
+
+    def pkg_label_renderer_data_func(self, cell_layout, renderer, model, iter,
+                                     data):
+        data = model.get_value(iter, LIST_UPDATE_DATA)
+        name = GLib.markup_escape_text(model.get_value(iter, LIST_NAME))
+
+        if data.group:
+            markup = name
+        elif data.item:
+            markup = name
+        else:  # header
+            markup = "<b>%s</b>" % name
+
+        renderer.set_property("markup", markup)
 
     def set_changes_buffer(self, changes_buffer, text, name, srcpkg):
         changes_buffer.set_text("")
@@ -264,16 +492,21 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
         iter = model.get_iter(path)
 
         # set descr
-        pkg = model.get_value(iter, LIST_PKG)
-        if (pkg is None or pkg.candidate is None or
-                pkg.candidate.description is None):
+        data = model.get_value(iter, LIST_UPDATE_DATA)
+        item = data.item
+        if (item is None and data.group is not None and
+                data.group.core_item is not None):
+            item = data.group.core_item
+        if (item is None or item.pkg is None or 
+                item.pkg.candidate is None or
+                item.pkg.candidate.description is None):
             changes_buffer = self.textview_changes.get_buffer()
             changes_buffer.set_text("")
             desc_buffer = self.textview_descr.get_buffer()
             desc_buffer.set_text("")
             self.notebook_details.set_sensitive(False)
             return
-        long_desc = pkg.candidate.description
+        long_desc = item.pkg.candidate.description
         self.notebook_details.set_sensitive(True)
         # do some regular expression magic on the description
         # Add a newline before each bullet
@@ -290,7 +523,7 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
         desc_buffer.set_text(long_desc)
 
         # now do the changelog
-        name = model.get_value(iter, LIST_NAME)
+        name = item.pkg.name
         if name is None:
             return
 
@@ -334,11 +567,10 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
                 button.disconnect(id)
         # check if we still are in the right pkg (the download may have taken
         # some time and the user may have clicked on a new pkg)
-        path = widget.get_cursor()[0]
-        if path is None:
+        now_path = widget.get_cursor()[0]
+        if now_path is None:
             return
-        now_name = widget.get_model()[path][LIST_NAME]
-        if name != now_name:
+        if path != now_path:
             return
         # display NEWS.Debian first, then the changelog
         changes = ""
@@ -350,7 +582,7 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
         if changes:
             self.set_changes_buffer(changes_buffer, changes, name, srcpkg)
 
-    def show_context_menu(self, widget, event):
+    def on_treeview_button_press(self, widget, event):
         """
         Show a context menu if a right click was performed on an update entry
         """
@@ -361,10 +593,13 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
             self.menu = menu = Gtk.Menu()
             item_select_none = \
                 Gtk.MenuItem.new_with_mnemonic(_("_Deselect All"))
-            item_select_none.connect("activate", self.select_none_updgrades)
+            item_select_none.connect("activate", self.select_none_upgrades)
             menu.append(item_select_none)
+            num_updates = self.cache.install_count
+            if num_updates == 0:
+                item_select_none.set_property("sensitive", False)
             item_select_all = Gtk.MenuItem.new_with_mnemonic(_("Select _All"))
-            item_select_all.connect("activate", self.select_all_updgrades)
+            item_select_all.connect("activate", self.select_all_upgrades)
             menu.append(item_select_all)
             menu.show_all()
             menu.popup_for_device(
@@ -373,35 +608,36 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
             return True
 
     # we need this for select all/unselect all
-    def _toggle_origin_headers(self, new_selection_value):
-        """ small helper that will set/unset the origin headers
+    def _toggle_group_headers(self, new_selection_value):
+        """ small helper that will set/unset the group headers
         """
         model = self.treeview_update.get_model()
         for row in model:
-            if not model.get_value(row.iter, LIST_PKG):
-                model.set_value(row.iter, LIST_TOGGLE_CHECKED,
+            data = model.get_value(row.iter, LIST_UPDATE_DATA)
+            if data.groups is not None or data.group is not None:
+                model.set_value(row.iter, LIST_TOGGLE_ACTIVE,
                                 new_selection_value)
 
-    def select_all_updgrades(self, widget):
+    def select_all_upgrades(self, widget):
         """
         Select all updates
         """
         self.setBusy(True)
         self.cache.saveDistUpgrade()
-        self._toggle_origin_headers(True)
+        self._toggle_group_headers(True)
         self.treeview_update.queue_draw()
-        self.refresh_updates_count()
+        self.updates_changed()
         self.setBusy(False)
 
-    def select_none_updgrades(self, widget):
+    def select_none_upgrades(self, widget):
         """
         Select none updates
         """
         self.setBusy(True)
         self.cache.clear()
-        self._toggle_origin_headers(False)
+        self._toggle_group_headers(False)
         self.treeview_update.queue_draw()
-        self.refresh_updates_count()
+        self.updates_changed()
         self.setBusy(False)
 
     def setBusy(self, flag):
@@ -417,7 +653,21 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
         while Gtk.events_pending():
             Gtk.main_iteration()
 
-    def refresh_updates_count(self):
+    def _mark_selected_updates(self):
+        def foreach_cb(model, path, iter, data):
+            data = model.get_value(iter, LIST_UPDATE_DATA)
+            active = False
+            if data.item:
+                active = data.item.pkg.marked_install or \
+                         data.item.pkg.marked_upgrade
+            elif data.group:
+                active = data.group.packages_are_selected()
+            elif data.groups:
+                active = any([g.packages_are_selected() for g in data.groups])
+            model.set_value(iter, LIST_TOGGLE_ACTIVE, active)
+        self.store.foreach(foreach_cb, None)
+
+    def _refresh_updates_count(self):
         self.button_install.set_sensitive(self.cache.install_count)
         try:
             inst_count = self.cache.install_count
@@ -458,17 +708,20 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
             self.hbox_downsize.show()
             self.vbox_alerts.show()
 
+    def updates_changed(self):
+        self._mark_selected_updates()
+        self._refresh_updates_count()
+
     def update_count(self):
         """activate or disable widgets and show dialog texts correspoding to
            the number of available updates"""
-        self.refresh_updates_count()
+        self.updates_changed()
 
         text_header = None
         text_desc = None
 
         if self.custom_header is not None:
             text_header = self.custom_header
-            text_desc = self.custom_desc
         # show different text on first run (UX team suggestion)
         elif self.settings.get_boolean("first-run"):
             flavor = self.window_main.meta_release.flavor_name
@@ -481,11 +734,12 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
             text_header = _("Updated software is available for this "
                             "computer. Do you want to install it now?")
 
+        if self.custom_desc is not None:
+            text_desc = self.custom_desc
+
         self.notebook_details.set_sensitive(True)
         self.treeview_update.set_sensitive(True)
         self.button_install.grab_default()
-        self.treeview_update.set_cursor(Gtk.TreePath.new_from_string("1"),
-                                        None, False)
         self.label_header.set_label(text_header)
 
         if text_desc is not None:
@@ -556,7 +810,7 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
             self.cache.checkFreeSpace()
         except NotEnoughFreeSpaceError as e:
             for req in e.free_space_required_list:
-                self.window_main.start_error(err_sum,
+                self.window_main.start_error(False, err_sum,
                                              err_long % (req.size_total,
                                                          req.dir,
                                                          req.size_needed,
@@ -571,14 +825,14 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
         # can deal with dialup connections properly
         if state in NetworkManagerHelper.NM_STATE_CONNECTING_LIST:
             self.label_offline.set_text(_("Connecting..."))
-            self.refresh_updates_count()
+            self.updates_changed()
             self.hbox_offline.show()
             self.vbox_alerts.show()
             self.connected = False
         # in doubt (STATE_UNKNOWN), assume connected
         elif (state in NetworkManagerHelper.NM_STATE_CONNECTED_LIST or
               state == NetworkManagerHelper.NM_STATE_UNKNOWN):
-            self.refresh_updates_count()
+            self.updates_changed()
             self.hbox_offline.hide()
             self.connected = True
             # trigger re-showing the current app to get changelog info (if
@@ -588,7 +842,7 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
             self.connected = False
             self.label_offline.set_text(_("You may not be able to check for "
                                           "updates or download new updates."))
-            self.refresh_updates_count()
+            self.updates_changed()
             self.hbox_offline.show()
             self.vbox_alerts.show()
 
@@ -611,63 +865,37 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
             self.hbox_on_3g.hide()
             self.hbox_roaming.hide()
 
-    def row_activated(self, treeview, path, column):
-        iter = self.store.get_iter(path)
-
-        pkg = self.store.get_value(iter, LIST_PKG)
-        origin = self.store.get_value(iter, LIST_ORIGIN)
-        if pkg is not None:
-            return
-        self.toggle_from_origin(pkg, origin, True)
-
-    def toggle_from_origin(self, pkg, origin, select_all=True):
-        self.setBusy(True)
-        actiongroup = apt_pkg.ActionGroup(self.cache._depcache)
-        for pkg in self.list.pkgs[origin]:
-            if pkg.marked_install or pkg.marked_upgrade:
-                #print("marking keep: ", pkg.name)
-                pkg.mark_keep()
-            elif not (pkg.name in self.list.held_back):
-                #print("marking install: ", pkg.name)
-                pkg.mark_install(auto_fix=False, auto_inst=False)
-        # check if we left breakage
-        if self.cache._depcache.broken_count:
-            Fix = apt_pkg.ProblemResolver(self.cache._depcache)
-            Fix.resolve_by_keep()
-        self.refresh_updates_count()
-        self.treeview_update.queue_draw()
-        del actiongroup
-        self.setBusy(False)
-
-    def toggled(self, renderer, path):
+    def on_update_toggled(self, renderer, path):
         """ a toggle button in the listview was toggled """
         iter = self.store.get_iter(path)
-        pkg = self.store.get_value(iter, LIST_PKG)
-        origin = self.store.get_value(iter, LIST_ORIGIN)
+        data = self.store.get_value(iter, LIST_UPDATE_DATA)
         # make sure that we don't allow to toggle deactivated updates
         # this is needed for the call by the row activation callback
-        if pkg is None:
-            toggled_value = not self.store.get_value(iter, LIST_TOGGLE_CHECKED)
-            self.toggle_from_origin(pkg, origin, toggled_value)
-            self.store.set_value(iter, LIST_TOGGLE_CHECKED, toggled_value)
+        if data.groups or data.group:
+            if data.groups:
+                self.toggle_from_groups(data.groups)
+            elif data.group:
+                self.toggle_from_groups([data.group])
+            self.updates_changed()
             self.treeview_update.queue_draw()
             return
-        if pkg is None or pkg.name in self.list.held_back:
+        if (data.item is None or data.item.pkg is None or
+                data.item.pkg.name in self.list.held_back):
             return False
         self.setBusy(True)
         # update the cache
-        if pkg.marked_install or pkg.marked_upgrade:
-            pkg.mark_keep()
+        if data.item.pkg.marked_install or data.item.pkg.marked_upgrade:
+            data.item.pkg.mark_keep()
             if self.cache._depcache.broken_count:
                 Fix = apt_pkg.ProblemResolver(self.cache._depcache)
                 Fix.resolve_by_keep()
         else:
             try:
-                pkg.mark_install()
+                data.item.pkg.mark_install()
             except SystemError:
                 pass
+        self.updates_changed()
         self.treeview_update.queue_draw()
-        self.refresh_updates_count()
         self.setBusy(False)
 
     def on_treeview_update_row_activated(self, treeview, path, column, *args):
@@ -675,7 +903,29 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
         If an update row was activated (by pressing space), toggle the
         install check box
         """
-        self.toggled(None, path)
+        self.on_update_toggled(None, path)
+
+    def toggle_from_groups(self, groups):
+        self.setBusy(True)
+        actiongroup = apt_pkg.ActionGroup(self.cache._depcache)
+
+        # Deselect all updates if any in group are selected
+        keep_packages = any([g.packages_are_selected() for g in groups])
+        for group in groups:
+            for item in group.items:
+                if keep_packages:
+                    item.pkg.mark_keep()
+                elif not item.pkg.name in self.list.held_back:
+                    item.pkg.mark_install(auto_fix=False, auto_inst=False)
+
+        # check if we left breakage
+        if self.cache._depcache.broken_count:
+            Fix = apt_pkg.ProblemResolver(self.cache._depcache)
+            Fix.resolve_by_keep()
+        self.updates_changed()
+        self.treeview_update.queue_draw()
+        del actiongroup
+        self.setBusy(False)
 
     def save_state(self):
         """ save the state  (window-size for now) """
@@ -691,9 +941,58 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
         expanded = self.expander_details.get_expanded()
         self.window_main.set_resizable(expanded)
         if w > 0 and h > 0 and expanded:
-            self.window_main.resize(w, h)
+            # There is a race here.  If we immediately resize, it often doesn't
+            # take.  Using idle_add helps, but we *still* occasionally don't
+            # restore the size correctly.  Help needed to track this down!
+            GLib.idle_add(lambda: self.window_main.resize(w, h))
         return False
 
+    def _add_header(self, name, groups):
+        total_size = 0
+        for group in groups:
+            total_size = total_size + group.get_total_size()
+        header_row = [
+            name,
+            UpdateData(groups, None, None),
+            humanize_size(total_size),
+            True
+        ]
+        return self.store.append(None, header_row)
+
+    def _add_groups(self, groups):
+        # Each row contains:
+        #  row label (for screen reader),
+        #  update data tuple (is_toplevel, group object, package object),
+        #  update size,
+        #  update selection state
+        for group in groups:
+            if not group.items:
+                continue
+
+            group_is_item = None
+            if not isinstance(group, UpdateSystemGroup) and \
+                    len(group.items) == 1:
+                group_is_item = group.items[0]
+
+            group_row = [
+                group.name,
+                UpdateData(None, group, group_is_item),
+                humanize_size(group.get_total_size()),
+                True
+            ]
+            group_iter = self.store.append(None, group_row)
+
+            if group_is_item:
+                continue
+            for item in group.items:
+                item_row = [
+                    item.name,
+                    UpdateData(None, None, item),
+                    humanize_size(getattr(item.pkg.candidate, "size", 0)),
+                    True
+                ]
+                self.store.append(group_iter, item_row)
+
     def fillstore(self):
         # use the watch cursor
         self.setBusy(True)
@@ -704,46 +1003,22 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
         self.dl_size = 0
 
         self.scrolledwindow_update.show()
-        origin_list = sorted(
-            self.list.pkgs, key=operator.attrgetter("importance"),
-            reverse=True)
-        for origin in origin_list:
-            self.store.append(['<b><big>%s</big></b>' % origin.description,
-                               origin.description, None, origin, True])
-            for pkg in self.list.pkgs[origin]:
-                name = xml.sax.saxutils.escape(pkg.name)
-                if not pkg.is_installed:
-                    name += _(" (New install)")
-                summary = xml.sax.saxutils.escape(getattr(pkg.candidate,
-                                                          "summary", ""))
-                if self.summary_before_name:
-                    contents = "%s\n<small>%s</small>" % (summary, name)
-                else:
-                    contents = "<b>%s</b>\n<small>%s</small>" % (name, summary)
-                #TRANSLATORS: the b stands for Bytes
-                size = _("(Size: %s)") % humanize_size(getattr(pkg.candidate,
-                                                               "size", 0))
-                installed_version = getattr(pkg.installed, "version", None)
-                candidate_version = getattr(pkg.candidate, "version", None)
-                if installed_version is not None:
-                    version = _("From version %(old_version)s "
-                                "to %(new_version)s") % {
-                                    "old_version": installed_version,
-                                    "new_version": candidate_version}
-                else:
-                    version = _("Version %s") % candidate_version
-                if self.show_versions:
-                    contents = "%s\n<small>%s %s</small>" % (contents, version,
-                                                             size)
-                else:
-                    contents = "%s <small>%s</small>" % (contents, size)
-                self.store.append([contents, pkg.name, pkg, None, True])
+
+        # add security and update groups to self.store
+        if self.list.security_groups:
+            self._add_header(_("Security updates"), self.list.security_groups)
+            self._add_groups(self.list.security_groups)
+        if self.list.security_groups and self.list.update_groups:
+            self._add_header(_("Other updates"), self.list.update_groups)
+        if self.list.update_groups:
+            self._add_groups(self.list.update_groups)
+
         self.treeview_update.set_model(self.store)
         self.update_count()
         self.setBusy(False)
         while Gtk.events_pending():
             Gtk.main_iteration()
-        self.refresh_updates_count()
+        self.updates_changed()
         return False
 
     def main(self):
