@@ -13,12 +13,13 @@ from defer import inline_callbacks
 from aptdaemon.gtk3widgets import (AptCancelButton,
                                    AptConfigFileConflictDialog,
                                    AptDetailsExpander,
-                                   AptErrorDialog,
                                    AptMediumRequiredDialog,
                                    AptProgressBar)
 from aptdaemon.enums import (EXIT_SUCCESS,
                              EXIT_FAILED,
                              STATUS_COMMITTING,
+                             get_error_description_from_enum,
+                             get_error_string_from_enum,
                              get_status_string_from_enum)
 
 from UpdateManager.backend import InstallBackend
@@ -62,9 +63,9 @@ class InstallBackendAptdaemon(InstallBackend):
                                       _("Checking for updates…"),
                                       False, False)
         except errors.NotAuthorizedError:
-            self.emit("action-done", self.UPDATE, False, False)
+            self.emit("action-done", self.UPDATE, False, False, None, None)
         except:
-            self.emit("action-done", self.UPDATE, True, False)
+            self.emit("action-done", self.UPDATE, True, False, None, None)
             raise
 
     @inline_callbacks
@@ -84,14 +85,14 @@ class InstallBackendAptdaemon(InstallBackend):
                                       _("Installing updates…"),
                                       True, close_on_done)
         except errors.NotAuthorizedError as e:
-            self.emit("action-done", self.INSTALL, False, False)
+            self.emit("action-done", self.INSTALL, False, False, None, None)
         except dbus.DBusException as e:
             #print(e, e.get_dbus_name())
             if e.get_dbus_name() != "org.freedesktop.DBus.Error.NoReply":
                 raise
-            self.emit("action-done", self.INSTALL, False, False)
+            self.emit("action-done", self.INSTALL, False, False, None, None)
         except Exception as e:
-            self.emit("action-done", self.INSTALL, True, False)
+            self.emit("action-done", self.INSTALL, True, False, None, None)
             raise
 
     def _on_progress_changed(self, trans, progress):
@@ -153,6 +154,7 @@ class InstallBackendAptdaemon(InstallBackend):
         trans.connect("medium-required", self._on_medium_required)
         trans.connect("config-file-conflict", self._on_config_file_conflict)
 
+        yield trans.set_debconf_frontend("gnome")
         yield trans.run()
 
         self.window_main.push(builder.get_object("pane_update_progress"),
@@ -223,16 +225,17 @@ class InstallBackendAptdaemon(InstallBackend):
             transaction.resolve_config_file_conflict(old, "keep")
 
     def _on_finished(self, trans, status, action, close_on_done):
+        error_string = None
+        error_desc = None
         if status == EXIT_FAILED:
-            err_dia = AptErrorDialog(trans.error, self.window_main)
-            err_dia.run()
-            err_dia.hide()
-            sys.exit(0)
+            error_string = get_error_string_from_enum(trans.error.code)
+            error_desc = get_error_description_from_enum(trans.error.code)
         elif status == EXIT_SUCCESS and close_on_done:
             sys.exit(0)
         # tell unity to hide the progress again
         self.unity.set_progress(-1)
-        self.emit("action-done", action, True, status == EXIT_SUCCESS)
+        self.emit("action-done", action, True, status == EXIT_SUCCESS,
+                  error_string, error_desc)
 
 if __name__ == "__main__":
     b = InstallBackendAptdaemon(None)
