@@ -42,7 +42,8 @@ class InstallBackendSynaptic(InstallBackend):
         action, tempf = data
         if tempf:
             tempf.close()
-        self.emit("action-done", action, True, os.WEXITSTATUS(condition) == 0)
+        self.emit("action-done", action, True, os.WEXITSTATUS(condition) == 0,
+                  None, None)
 
     def update(self):
         opt = ["--update-at-startup"]
